#include <QFile>
#include "filesystem/NCFilesystem.hxx"
#include "filesystem/FileIdentifier.hxx"

// --- internal typedefs ---

using IdentifyType = NC::Identify::Type;
using ColType = NC::FileData::Type;

// --- internal objects ---

FileIdentifier id;

// --- constructors and deconstructors ---

NCFilesystem::NCFilesystem(QObject *parent)
: QObject(parent)
{
    emit sigFileStatus(0, _files.count(), ColType::None);
}

NCFilesystem::NCFilesystem(const NCFilesystem& rhs)
: QObject(rhs.parent()), _files(rhs._files)
{
    emit sigFileStatus(0, _files.count(), ColType::None);
}

NCFilesystem::NCFilesystem(const NCFilesystem&& rhs)
: QObject(std::move(rhs.parent())), _files(std::move(rhs._files))
{
    emit sigFileStatus(0, _files.count(), ColType::None);
}

NCFilesystem::~NCFilesystem()
{
    _files.clear();
    emit sigFileStatus(0, _files.count(), ColType::None);
}

// --- operators ---

NCFilesystem& NCFilesystem::operator=(const NCFilesystem& rhs)
{
    if (this != &rhs)
    {
        _files = rhs._files;
        emit sigFileStatus(0, _files.count(), ColType::All);
    }

    return *this;
}

NCFilesystem& NCFilesystem::operator=(const NCFilesystem&& rhs)
{
    if (this != &rhs)
    {
        _files = std::move(rhs._files);
        emit sigFileStatus(0, _files.count(), ColType::All);
    }

    return *this;
}

bool NCFilesystem::operator==(const NCFilesystem& rhs) const
{
    return _files == rhs._files;
}

bool NCFilesystem::operator!=(const NCFilesystem& rhs) const
{
    return _files != rhs._files;
}

// --- basic and status methodes ---

quint32 NCFilesystem::count() const
{
    return _files.count();
}

quint32 NCFilesystem::size() const
{
    quint32 result = HSize;

    for (auto& f : _files)
        result += f.size();

    return result;
}

bool NCFilesystem::valid() const
{
    auto it = _files.begin();
    bool result = true;

    while (result && it != _files.end())
    {
        result = result && (*it).valid();
        ++it;
    }

    return result;
}

void NCFilesystem::clear()
{
    _files.clear();
    emit sigFileStatus(0, _files.count(), ColType::None);
}

QString NCFilesystem::info() const
{
    QString result;

    for (qint32 i = 0; i < _files.count(); ++i)
    {
        result.append("idx:").append(QString::number(i));
        result.append(" hs:").append(QString::number(_files[i].hSize()));
        result.append(" pos:").append(QString::number(_files[i].position()));
        result.append(" os:").append(QString::number(_files[i].oSize()));
        result.append(" cs:").append(QString::number(_files[i].cSize()));
        result.append(" ln:").append(QString::number(_files[i].nameLength()));
        result.append(" name:'").append(_files[i].name()).append("'").append("\n");
    }

    return result;
}

IdentifyType NCFilesystem::fileIdentify(const QString& filename) const
{
    return id.identify(filename);
}

QString NCFilesystem::fileIdentifyName(const IdentifyType& type) const
{
    return id.identifyName(type);
}

QString NCFilesystem::fileIdentifyName(const QString& filename) const
{
    return id.identifyName(fileIdentify(filename));
}

quint32 NCFilesystem::load(const QString& filename)
{
    if (fileIdentify(filename) == id.VFS)
        return vfsLoader(filename, false);

    return Error;
}

quint32 NCFilesystem::save(const QString& filename) const
{
    return vfsSaver(filename);
}

quint32 NCFilesystem::fileLoad(const QString& filename)
{
    switch (fileIdentify(filename))
    {
        case id.VFS:
            return vfsLoader(filename, false);
        case id.PAK1:
            return pakLoader(filename, false);
        case id.Plain:
            return plainLoader(filename, false);
        case id.ZLIB:
            return zlibLoader(filename, false);
        default:
            break;
    }

    return Error;
}

quint32 NCFilesystem::fileAdd(const QString& filename)
{
    switch (fileIdentify(filename))
    {
        case id.VFS:
            return vfsLoader(filename, true);
        case id.PAK1:
            return pakLoader(filename, true);
        case id.Plain:
            return plainLoader(filename, true);
        case id.ZLIB:
            return zlibLoader(filename, true);
        default:
            break;
    }

    return Error;
}

quint32 NCFilesystem::fileAdd(const NCFile& entry)
{
    _files.push_back(entry);
    recalc(_files);

    emit sigFileAdd(_files.back(), _files.count() - 1);
    emit sigFileStatus(0, _files.count(), ColType::Position);

    return _files.size();
}

quint32 NCFilesystem::fileSave(const QString& filename, const QString& name) const
{
    return plainSaver(fileGet(name), filename);
}

quint32 NCFilesystem::fileSave(const QString& filename, const quint32& index) const
{
    return plainSaver(fileGet(index), filename);
}

quint32 NCFilesystem::fileSave(const QString& filename, const QString& name, const IdentifyType& type) const
{
    switch (type)
    {
        case id.PAK1:
            return pakSaver(fileGet(name), filename);
        case id.Plain:
            return plainSaver(fileGet(name), filename);
        case id.ZLIB:
            return zlibSaver(fileGet(name), filename);
        default:
            break;
    }

    return Error;
}

quint32 NCFilesystem::fileSave(const QString& filename, const quint32& index, const IdentifyType& type) const
{
    switch (type)
    {
        case id.PAK1:
            return pakSaver(fileGet(index), filename);
        case id.Plain:
            return plainSaver(fileGet(index), filename);
        case id.ZLIB:
            return zlibSaver(fileGet(index), filename);
        default:
            break;
    }

    return Error;
}

NCFile NCFilesystem::fileGet(const QString& name) const
{
    NCFile result;

    for (auto& f : _files)
    {
        if (name == f.name())
        {
            result = f;
            break;
        }
    }

    return result;
}

NCFile NCFilesystem::fileGet(const quint32& index) const
{
    NCFile result;

    if (index < static_cast<quint32>(_files.count()))
        result = _files[index];

    return result;
}

NCFile NCFilesystem::fileReplace(const NCFile& entry, const QString& name)
{
    NCFile result;

    if (!entry.valid())
        return result;

    for (auto& f : _files)
    {
        if (name == f.name())
        {
            result = f;
            f = entry;

            emit sigFileReplace(f, _files.indexOf(f));
            emit sigFileStatus(0, _files.count(), ColType::Position);

            break;
        }
    }
    recalc(_files);

    return result;
}

NCFile NCFilesystem::fileReplace(const NCFile& entry, const quint32& index)
{
    NCFile result;

    if (!entry.valid())
        return result;

    if (index >= static_cast<quint32>(_files.count()))
        return result;

    result = _files[index];
    _files[index] = entry;
    recalc(_files);

    emit sigFileReplace(_files[index], index);
    emit sigFileStatus(0, _files.count(), ColType::Position);

    return result;
}

NCFile NCFilesystem::fileReplace(const QString& filename, const QString& name)
{
    NCFile result;

    if (filename.isEmpty())
        return result;

    for (auto& f : _files)
    {
        if (name == f.name())
        {
            NCFilesystem temp;

            temp.fileLoad(filename);
            if (temp.count())
            {
                const quint32 index = _files.indexOf(f);
                const quint32 added = temp.count();

                result = f;
                _files.remove(index);
                while (temp.count())
                {
                    _files.insert(index, temp.fileGet(temp.count() - 1));
                    temp.fileRemove(temp.count() - 1);
                }
                recalc(_files);

                emit sigFileReplace(_files[index], index);
                for (quint32 i = index; i < (index + added); ++i)
                    emit sigFileAdd(_files[i], i);
                emit sigFileStatus(0, _files.count(), ColType::Position);
            }
            break;
        }
    }

    return result;
}

NCFile NCFilesystem::fileReplace(const QString& filename, const quint32& index)
{
    NCFile result;
    NCFilesystem temp;

    if (filename.isEmpty())
        return result;

    if (index >= static_cast<quint32>(_files.count()))
        return result;

    temp.fileLoad(filename);
    if (temp.count())
    {
        const quint32 added = temp.count();

        result = _files[index];
        _files.remove(index);
        while (temp.count())
        {
            _files.insert(index, temp.fileGet(temp.count() - 1));
            temp.fileRemove(temp.count() - 1);
        }
        recalc(_files);

        emit sigFileReplace(_files[index], index);
        for (quint32 i = (index + 1); i < (index + added); ++i)
            emit sigFileAdd(_files[i], i);
        emit sigFileStatus(0, _files.count(), ColType::Position);
    }

    return result;
}

NCFile NCFilesystem::fileRemove(const QString& name)
{
    NCFile result;

    if (name.isEmpty())
        return result;

    for (auto it = _files.begin(); it != _files.end(); ++it)
    {
        if (name == (*it).name())
        {
            emit sigFileRemove(_files.indexOf(it));
            emit sigFileStatus(0, _files.count(), ColType::Position);

            result = *it;
            _files.erase(it);
            break;
        }
    }
    recalc(_files);

    return result;
}

NCFile NCFilesystem::fileRemove(const quint32& index)
{
    NCFile result;

    if (index >= static_cast<quint32>(_files.count()))
        return result;

    result = _files[index];
    _files.erase(_files.begin() + index);
    recalc(_files);

    emit sigFileRemove(index);
    emit sigFileStatus(0, _files.count(), ColType::Position);

    return result;
}

// --- internal methodes ---

void NCFilesystem::recalc(QVector<NCFile>& files)
{
    quint32 base_offset = sizeof (id.VFS) + sizeof (quint32);
    quint32 csize_offset = 0;

    for (auto& f : files)
        base_offset += f.headerSize() + f.nameLength();

    for (auto& f : files)
    {
        f.setPosition(base_offset + csize_offset);
        csize_offset += f.cSize();
    }
}

quint32 NCFilesystem::vfsLoader(const QString& filename, const bool add)
{
    QFile ifile(filename);

    if (ifile.open(QIODevice::ReadOnly))
    {
        QVector<NCFile> files;
        const quint32 old_size = _files.count();
        quint32 data = 0;
        quint32 count = 0;
        QByteArray stream;

        ifile.seek(sizeof (id.VFS));
        ifile.read(reinterpret_cast<char *>(&count), sizeof (count));

        for (quint32 i = 0; i < count; ++i)
        {
            NCFile entry;
            quint32 pos = 0;

            // header + name
            ifile.read(reinterpret_cast<char *>(&data), sizeof (data));
            entry.setHSize(data);
            ifile.read(reinterpret_cast<char *>(&data), sizeof (data));
            entry.setPosition(data);
            ifile.read(reinterpret_cast<char *>(&data), sizeof (data));
            entry.setCSize(data);
            ifile.read(reinterpret_cast<char *>(&data), sizeof (data));
            entry.setOSize(data);
            ifile.read(reinterpret_cast<char *>(&data), sizeof (data));
            entry.setNameLength(data);
            stream.resize(data);
            ifile.read(stream.data(), data);
            stream.truncate(data);
            entry.setName(stream);
            // compressed data
            pos = ifile.pos();
            ifile.seek(entry.position());
            stream.resize(entry.cSize());
            ifile.read(stream.data(), entry.cSize());
            entry.setCData(stream);
            ifile.seek(pos);

            files.push_back(entry);
        }
        ifile.close();

        if (add)
        {
            _files += files;
            recalc(_files);

            emit sigFileStatus(0, _files.count(), ColType::Position);
            emit sigFileStatus(old_size, _files.count(), ColType::All);
        }
        else
        {
            recalc(files);
            _files = files;

            emit sigFileStatus(0, _files.count(), ColType::None);
        }

        return _files.count();
    }

    return Error;
}

quint32 NCFilesystem::pakLoader(const QString& filename, const bool add)
{
    const quint32 IDSize = sizeof (id.PAK1) + sizeof (id.PAK2) + sizeof (id.PAK3);
    QFile ifile(filename);

    if (ifile.open(QIODevice::ReadOnly))
    {
        quint32 data = 0;
        QByteArray stream;
        NCFile entry;

        // header data
        entry.setPosition(0);
        entry.setCSize(ifile.size() - IDSize);
        ifile.seek(IDSize);
        ifile.read(reinterpret_cast<char *>(&data), sizeof (data));
        entry.setOSize(data);
        entry.setNameLength(filename.size() + 1);
        entry.setName(filename);
        entry.setHSize(entry.headerSize() + entry.nameLength());
        // compressed data
        stream.resize(entry.cSize());
        ifile.read(stream.data(), entry.cSize());
        entry.setCData(stream);
        ifile.close();

        if (!add)
            _files.clear();

        _files.push_back(entry);
        recalc(_files);

        emit sigFileAdd(_files.back(), _files.count() - 1);
        emit sigFileStatus(0, _files.count(), ColType::Position);

        return _files.count();
    }

    return Error;
}

quint32 NCFilesystem::plainLoader(const QString& filename, const bool add)
{
    QFile ifile(filename);

    if (ifile.open(QIODevice::ReadOnly))
    {
        QByteArray stream;
        NCFile entry;

        stream.resize(ifile.size());
        ifile.read(stream.data(), ifile.size());
        entry.setOData(stream); // odata compresses stream to cdata
        entry.setPosition(0);
        entry.setCSize(entry.cData().size());
        entry.setOSize(ifile.size());
        entry.setNameLength(filename.size() + 1);
        entry.setName(filename);
        entry.setHSize(entry.headerSize() + entry.nameLength());
        ifile.close();

        if (!add)
            _files.clear();

        _files.push_back(entry);
        recalc(_files);

        emit sigFileAdd(_files.back(), _files.count() - 1);
        emit sigFileStatus(0, _files.count(), ColType::Position);

        return _files.count();
    }

    return Error;
}

quint32 NCFilesystem::zlibLoader(const QString& filename, const bool add)
{
    QFile ifile(filename);

    if (ifile.open(QIODevice::ReadOnly))
    {
        QByteArray stream;
        NCFile entry;

        stream.resize(ifile.size());
        ifile.read(stream.data(), ifile.size());
        entry.setCData(stream);
        entry.setPosition(0);
        entry.setCSize(ifile.size());
        entry.setOSize(entry.oData().size());
        entry.setNameLength(filename.size() + 1);
        entry.setName(filename);
        entry.setHSize(entry.headerSize() + entry.nameLength());
        ifile.close();

        if (!add)
            _files.clear();

        _files.push_back(entry);
        recalc(_files);

        emit sigFileAdd(_files.back(), _files.count() - 1);
        emit sigFileStatus(0, _files.count(), ColType::Position);

        return _files.count();
    }

    return Error;
}

quint32 NCFilesystem::vfsSaver(const QString& filename) const
{
    QFile ofile(filename);

    if (!valid())
        return Error;

    if (ofile.open(QIODevice::WriteOnly))
    {
        quint32 data = 0;
        const quint32 count = _files.count();

        data = id.VFS;
        ofile.write(reinterpret_cast<const char *>(&data), sizeof (data));
        ofile.write(reinterpret_cast<const char *>(&count), sizeof (count));

        for (auto& f : _files)
        {
            data = f.hSize();
            ofile.write(reinterpret_cast<const char *>(&data), sizeof (data));
            data = f.position();
            ofile.write(reinterpret_cast<const char *>(&data), sizeof (data));
            data = f.cSize();
            ofile.write(reinterpret_cast<const char *>(&data), sizeof (data));
            data = f.oSize();
            ofile.write(reinterpret_cast<const char *>(&data), sizeof (data));
            data = f.nameLength();
            ofile.write(reinterpret_cast<const char *>(&data), sizeof (data));
            ofile.write(f.name().toStdString().c_str(), f.nameLength());
        }

        for (auto& f : _files)
            ofile.write(f.cData());

        ofile.close();

        emit sigFileStatus(0, _files.count(), ColType::None);

        return _files.count();
    }

    return Error;
}

quint32 NCFilesystem::pakSaver(const NCFile& entry, const QString& filename) const
{
    QFile ofile(filename);

    if (!entry.valid())
        return Error;

    if (ofile.open(QIODevice::WriteOnly))
    {
        quint32 data = 0;

        data = id.PAK1;
        ofile.write(reinterpret_cast<const char *>(&data), sizeof (data));
        data = id.PAK2;
        ofile.write(reinterpret_cast<const char *>(&data), sizeof (data));
        data = id.PAK3;
        ofile.write(reinterpret_cast<const char *>(&data), sizeof (data));
        data = entry.oSize();
        ofile.write(reinterpret_cast<const char *>(&data), sizeof (data));
        ofile.write(entry.cData());

        ofile.close();

        return 1;
    }

    return Error;
}

quint32 NCFilesystem::plainSaver(const NCFile& entry, const QString& filename) const
{
    QFile ofile(filename);

    if (!entry.valid())
        return Error;

    if (ofile.open(QIODevice::WriteOnly))
    {
        ofile.write(entry.oData());
        ofile.close();

        return 1;
    }

    return Error;
}

quint32 NCFilesystem::zlibSaver(const NCFile& entry, const QString& filename) const
{
    QFile ofile(filename);

    if (!entry.valid())
        return Error;

    if (ofile.open(QIODevice::WriteOnly))
    {
        ofile.write(entry.cData());
        ofile.close();

        return 1;
    }

    return Error;
}
